/******************************************************************************/
/*  Wireless Temp/Humidity Logger                                             */
/*  Transmitter unit main program                                             */
/*  Copyright(c) 2013 Hirofumi Momose  hirofumi[at]momose.com                 */
/******************************************************************************/

/******************************************************************************/
/* Files to Include                                                           */
/******************************************************************************/
#include <xc.h>         /* XC8 General Include File */
#include <stdint.h>        /* For uint8_t definition */
#include <stdbool.h>       /* For true/false definition */
#include <stdlib.h>

/******************************************************************************/
/* Configuration Bits (see P16F1827.INC)                                      */
/******************************************************************************/
// Config 1
#pragma config FOSC = INTOSC    // LP, XT, HS, EXTRC, INTOSC, ECL, ECM, ECH
#pragma config WDTE = SWDTEN    // OFF, SWDTEN, NSLEEP, ON
#pragma config PWRTE = OFF      // ON, OFF
#pragma config MCLRE = ON       // ON, OFF
#pragma config CP = OFF         // ON, OFF
#pragma config CPD = OFF        // ON, OFF
#pragma config BOREN = OFF      // OFF, SBODEN, NSLEEP, ON
#pragma config CLKOUTEN = OFF   // ON, OFF
#pragma config IESO = OFF       // ON, OFF
#pragma config FCMEN = OFF      // ON, OFF

// Config 2
//#pragma config WRT = OFF      // ALL, HALF, BOOT, OFF
//#pragma config PLLEN = OFF    // OFF, ON
//#pragma config STVREN = OFF   // OFF, ON
#pragma config BORV = HI        // HI, LO, 19(same as LO)
#pragma config LVP = OFF        // OFF, ON

//#pragma config DEBUG = ON
//#pragma CONFIG2 = 0x0000;

/******************************************************************************/
/* Pin definision / Macro control                                             */
/******************************************************************************/
#define LED_ON()        (LATAbits.LATA1 = 1)
#define LED_OFF()       (LATAbits.LATA1 = 0)

#define SENS_PWON()     {LATBbits.LATB1=1; LATBbits.LATB2=1;}
#define SENS_PWOFF()    {LATBbits.LATB1=0; LATBbits.LATB2=0;}

#define TX_PWON()       (LATAbits.LATA3 = 1)
#define TX_PWOFF()      (LATAbits.LATA3 = 0)
#define TX_KEYDN()      (LATAbits.LATA2 = 1)
#define TX_KEYUP()      (LATAbits.LATA2 = 0)

#define DCDC_ON()       (LATAbits.LATA4 = 0)    // DC-DC enable control
#define DCDC_OFF()      (LATAbits.LATA4 = 0)    // is not used.

/******************************************************************************/
/* Global Constants / Variable Declaration                                    */
/******************************************************************************/
#define SYS_FREQ_HI     8000000L
#define SYS_FREQ        500000L
#define FCY_HI          SYS_FREQ_HI/4
#define FCY             SYS_FREQ/4
#define _XTAL_FREQ      SYS_FREQ

#define TMR1CLK         (FCY)
#define TMRCLK_HI       (FCY_HI)

#define USEC2TMR1(x)    (TMR1CLK*(x)/1000000)
#define MSEC2TMR1(x)    (TMR1CLK*(x)/1000)
#define USEC2TMR1_HI(x) (TMRCLK_HI/1000000L*(x))
#define MSEC2TMR1_HI(x) (TMRCLK_HI/1000*(x))

/******************************************************************************/
/* Function Prototypes                                                        */
/******************************************************************************/
void send_data_int(void);
void send_data(union xdata_u *);
void read_sensor_int(void);
unsigned long read_sensor(void);
unsigned char read_vbatt(void);
#define DCDC_ENABLE     0x01
#define INTERRUPTIBLE   0x02
void die(void);
unsigned char low_power_sleep(unsigned long, unsigned char);

/******************************************************************************/
/* Interrupt Routines                                                         */
/******************************************************************************/
unsigned char rs_int = 0;
void interrupt isr(void)
{
    if(rs_int){
        read_sensor_int();
    } else {
        send_data_int();
    }
}

/******************************************************************************/
/* Data Xmitter                                                               */
/******************************************************************************/
#define BIT_PERIOD      (MSEC2TMR1(2))
#define CARR_PERIOD     (MSEC2TMR1(0))

unsigned char payload[10];
unsigned char cnt;
unsigned short pre, sfd;
unsigned char data;
enum xstate_e {
    xs_pre,
    xs_sfd,
    xs_pay,
    xs_fin
} xstate;

// Since XC8 can not handle 12bit-field data width
// Use 3 char data to carry 12bit temp and humidity.
union xdata_u {
    struct {
        unsigned char data[3];
        unsigned vbatt :4;
        unsigned id :4;
    };
    struct {
        unsigned char c1, c2, c3, c4;
    };
    unsigned long uldata;
};

void send_data_int(void)
{
    unsigned char x;

    PIR1bits.TMR1IF = 0;
    TMR1 += 0xffff - BIT_PERIOD;

    switch(xstate){

        case xs_pre:
            x = (unsigned char)((pre & 0x8000) >> 8);
            pre <<= 1;
            if(cnt == 15){
                cnt = 0;
                xstate = xs_sfd;
                break;
            }
            cnt++;
            break;

        case xs_sfd:
            x =  (unsigned char)((sfd & 0x8000) >> 8);
            sfd <<= 1;
            if(cnt == 15){
                cnt = 0;
                xstate = xs_pay;
                break;
            }
            cnt++;
            break;

        case xs_pay:
            //if(cnt %8 == 0){
            if((cnt & 0x07) == 0){
                // data = payload[cnt /8];
                data = payload[cnt >> 3];
            }
            x = data & 0x80;
            data <<= 1;
            if(cnt == 79){
                xstate = xs_fin;
                break;
            }
            cnt ++;
            break;

        case xs_fin:
            x = 0;
            PIE1bits.TMR1IE = 0;
            T1CONbits.TMR1ON = 0;
            break;
    }

    if(x){
        TX_KEYDN();
    } else {
        TX_KEYUP();
    }
}

void send_data(union xdata_u *datap)
{
    unsigned long uldata;
    unsigned char sum;
    unsigned char c, x, tmp;

    uldata = datap->uldata;

    for(tmp=0, c=0; c<32; c++){
        x = (unsigned char)((uldata & 0x80000000) >> 24);
        uldata <<= 1;
        if(x){
            tmp |= 0b01;
        } else {
            tmp |= 0b10;
        }
        if((c & 0x03) == 0x03){
            payload[c >>2] = tmp;
            tmp = 0;
        } else {
            tmp <<= 2;
        }
    }

    sum = datap->c1 + datap->c2 + datap->c3 + datap->c4;

    for(tmp =0, c=0; c<8; c++){
        x = sum & 0x80;
        sum <<= 1;
        if(x){
            tmp |= 0b01;
        } else {
            tmp |= 0b10;
        }
        if((c & 0x03) == 0x03){
            payload[8 + (c >> 2)] = tmp;
            tmp = 0;
        } else {
            tmp <<= 2;
        }
    }

    TX_PWON();
    LED_ON();
    __delay_ms(10);
    LED_OFF();
    __delay_ms(10);
    xstate = xs_pre;
    cnt = 0;
    pre = 0b1010101010101010;
    sfd = 0b1010101010101011;
    TX_KEYDN();                     // Send carrier for some extent of period
    PIR1bits.TMR1IF = 0;            // to stabilize AGC quickly
    PIE1bits.TMR1IE = 1;
    TMR1 = 0xffff - CARR_PERIOD;    
    T1CONbits.TMR1ON = 1;
    // sync
    while(xstate != xs_fin) continue;
    __delay_ms(10);
    TX_PWOFF();
}

/******************************************************************************/
/* Read sensor                                                                */
/******************************************************************************/
#define DATAHL_PERIOD   (USEC2TMR1_HI(100))

unsigned short tm_saved;
unsigned short sens_data[3];
unsigned char istimeout;
unsigned char bitcount;

union sdata_u {
    struct {
        short temp;
        short humid;
    };
    unsigned long uldata;
};

#define SDATA_INVALID (0xffffffff)

enum sstate_e {
    ss_req,
    ss_wait_ack1,
    ss_wait_ack2,
    ss_data,
    ss_fini
} sstate;

void read_sensor_int(void)
{
    unsigned char index;

    tm_saved = TMR1;
    TMR1 = 0;

    switch(sstate){

        case ss_req: // Int when 20ms elapsed
            sstate = ss_wait_ack1;
            PIR1bits.TMR1IF = 0;
            
            // Change Port direction to input
            TRISBbits.TRISB0 = 1;
            
            // Enable RB0 int 
            INTCONbits.INTF = 0;
            INTCONbits.INTE = 1;
            break;

        case ss_wait_ack1:
            INTCONbits.INTF = 0;
            if(PIR1bits.TMR1IF){
                // timeout error
                INTCONbits.INTE = 0;
                PIE1bits.TMR1IE = 0;
                T1CONbits.TMR1ON = 0;
                istimeout = 1;
                sstate = ss_fini;
                break;
            }
            sstate = ss_wait_ack2;
            bitcount = 0;
            break;

        case ss_wait_ack2:
            INTCONbits.INTF = 0;
            sstate = ss_data;
            break;

        case ss_data:
            INTCONbits.INTF = 0;
            if(PIR1bits.TMR1IF){
                // timeout error
                INTCONbits.INTE = 0;
                PIE1bits.TMR1IE = 0;
                T1CONbits.TMR1ON = 0;
                istimeout = 1;
                sstate = ss_fini;
                break;
            }
            index = (bitcount >> 4);
            sens_data[index] <<= 1;
            if(tm_saved > DATAHL_PERIOD){
                sens_data[index] |= 1;
            } else {
                sens_data[index] |= 0;
            }
            
            if(bitcount == 39){
                // Data recv done
                sstate = ss_fini;
                INTCONbits.INTE = 0;
                PIE1bits.TMR1IE = 0;
                T1CONbits.TMR1ON = 0;
                break;
            }
            bitcount++;
            break;

        case ss_fini:
            INTCONbits.INTF = 0;
            INTCONbits.INTE = 0;
            PIE1bits.TMR1IE = 0;
            PIR1bits.TMR1IF = 0;
            T1CONbits.TMR1ON = 0;
            break;
    }
}

unsigned long read_sensor(void)
{
    union sdata_u sdata; 
    unsigned char sum, sum2;

    SENS_PWON();
    // Sensor settling time > 1s required
    low_power_sleep(1200, DCDC_ENABLE);

    // Clock up to 8MHz
    OSCCONbits.IRCF = 0x0e;   // Select 8MHz
    OSCCONbits.SCS = 2;
    // Wait for Clock to stable
    while( OSCSTATbits.HFIOFL == 0) continue;
    // Now CPU running at 8MHz clock torr of within 2%

    sens_data[0] = 0;
    sens_data[1] = 0;
    sens_data[2] = 0;

    // Disable INT (not needed ?)
    INTCONbits.INTF = 0;
    INTCONbits.INTE = 0;

    istimeout = 0;
    sstate = ss_req;

    // Request data to sensor
    TRISBbits.TRISB0 = 0;   // Dir=Output
    LATBbits.LATB0 = 0;     // Make Low

    rs_int = 1;
    // Set timer int duration
    TMR1 = 0xffff - MSEC2TMR1_HI(1);

    // Enable timer INT
    PIR1bits.TMR1IF = 0;
    PIE1bits.TMR1IE = 1;
    T1CONbits.TMR1ON = 1;

    // Data recv will be processed in the ISR
    // Block main loop until data reception completed
    while(sstate != ss_fini ) continue;
    rs_int = 0;

    sdata.uldata = SDATA_INVALID;

    if(istimeout) goto fini;

    // Sum Check
    sum  = (sens_data[0] >> 8) & 0xff;
    sum +=  sens_data[0] & 0xff;
    sum += (sens_data[1] >> 8) & 0xff;
    sum +=  sens_data[1] & 0xff;

    sum2 = sens_data[2] & 0xff;
    if(sum != sum2){
        goto fini;
    }

    // Data output
    sdata.temp = sens_data[1] & 0x7fff;
    if(sens_data[1] & 0x8000){
        sdata.temp = -sdata.temp;
    }
    sdata.humid = sens_data[0];

fini:
    // Clock down to 500kHz
    OSCCONbits.IRCF = 7;
    while( OSCSTATbits.MFIOFR == 0) continue;

    SENS_PWOFF();
    return sdata.uldata;
}

/******************************************************************************/
/* Read battery voltage                                                       */
/******************************************************************************/
#define VREF    (3.3)
#define VOFS    (1.6)
#define VSHUT   (1.4)
unsigned char nshut = 0;
/*
 * AD reference is from VCC-VDD (3.3V typ)
 * return width is 4bit (0 to 15)
 * offset = 1.6V , LSB =0.1V
 * value  0 ... 1.6V
 * value 15 ... 3.1V
 */
unsigned char read_vbatt(void)
{
    unsigned char rc;

    ADCON0bits.ADON = 1;
    // __delay_us(10);

    // Start A/D
    ADCON0bits.GO_nDONE = 1;

    // Wait until completion
    while(ADCON0bits.GO_nDONE) continue;

    ADCON0bits.ADON = 0;

    if(ADRESH < 256*VSHUT/VREF){
        if(nshut > 10){
            die();
        }
        nshut++;
    } else {
        nshut = 0;
    }

    // Offset
    if(ADRESH > (unsigned char)(256*VOFS/VREF)){
        rc = ADRESH - (unsigned char)(256*VOFS/VREF);
    } else {
        rc = 0;
    }

    // Down to 4bit width
    rc = (unsigned char)((double)rc * 16/(256 - 256*VOFS/VREF) +0.5);
    if(rc > 15){
        rc = 15;
    }

    return rc;
}

/******************************************************************************/
/* Die (stop working until battery replacement. reduce leakage risk)          */
/******************************************************************************/
void die(void)
{
    TX_PWOFF();
    SENS_PWOFF();
    LED_OFF();
    INTCONbits.GIE = 0;
    asm("sleep");
}

/******************************************************************************/
/* Low power sleep                                                            */
/******************************************************************************/
unsigned long elapsed_ms;
unsigned char low_power_sleep(unsigned long sleep_ms, unsigned char flags)
{
    unsigned char rc = 0;
    elapsed_ms = 0;

    if(~flags & DCDC_ENABLE) DCDC_OFF();

    // Disable pull-ups to reduce power consumption during sleep.
    // Only PushSW (which can break this sleep)pull-up is kept.
    // Since PushSW is normally opened, current drawn for PushSW
    // does not matter.
    // Change DIPSW to output direction
    TRISBbits.TRISB4 = 0; // DIPSW
    TRISBbits.TRISB5 = 0; // DIPSW
    TRISBbits.TRISB6 = 0; // DIPSW
    TRISBbits.TRISB7 = 0; // DIPSW
    WPUB = 0x08;

    // Clear PushSW Interrupt on change flag bit.
    // This flag is polled during sleep-wakeup loop as below.
    // No actual interrupt is used.
    IOCBFbits.IOCBF3 = 0;

    // Set WDT duration
    WDTCONbits.WDTPS = 0b00111;     // 128ms
    //WDTCONbits.WDTPS = 0b01000;   // 256ms
    //WDTCONbits.WDTPS = 0b01001;   // 512ms
    //WDTCONbits.WDTPS = 0b01010;   // 1s
    //WDTCONbits.WDTPS = 0b01011;   // 2s interval (default value)
    WDTCONbits.SWDTEN = 1;  // Enable WDT

    while(elapsed_ms < sleep_ms){
        asm("sleep");
        if(IOCBFbits.IOCBF3 && (flags & INTERRUPTIBLE)){
            // PushSW pressed during sleep.
            rc = 1;
            break;
        }
        elapsed_ms += 128;          // Shoud match WDT duration
    }

    WDTCONbits.SWDTEN = 0;          // Disable WDT
    //OPTION_REGbits.nWPUEN = 0; // re-enable pull-up
    //INTCONbits.GIE = 1;
    // Change dipsw to input direction
    TRISBbits.TRISB4 = 1; // DIPSW
    TRISBbits.TRISB5 = 1; // DIPSW
    TRISBbits.TRISB6 = 1; // DIPSW
    TRISBbits.TRISB7 = 1; // DIPSW
    WPUB = 0xff;

    DCDC_ON();
    return rc;
}

/******************************************************************************/
/* Main                                                                       */
/******************************************************************************/
// Data sample interval in ms
#define SAMPLE_INTERVAL  60000
#define TXDELAY_MAX      48000

void main(void)
{
    union sdata_u sdata;
    union xdata_u xdata;
    unsigned long delay1, delay2;
    unsigned char id=0;
    unsigned char nforce = 0;

    // Clear port
    LATA = 0;
    LATB = 0;

    // Setup port direction
    // Output
    TRISAbits.TRISA1 = 0;   // LED
    TRISAbits.TRISA2 = 0;   // TX KEY
    TRISAbits.TRISA3 = 0;   // TX PWR
    TRISAbits.TRISA4 = 0;   // DC-DC ON
    TRISBbits.TRISB1 = 0;   // SENS PWR
    TRISBbits.TRISB2 = 0;   // SENS PWR
    TRISAbits.TRISA6 = 0;   // NC
    TRISAbits.TRISA7 = 0;   // NC

    // Input
    //TRISBbits.TRISB3 = 1; // PushSW
    //TRISBbits.TRISB4 = 1; // DIPSW
    //TRISBbits.TRISB5 = 1; // DIPSW
    //TRISBbits.TRISB6 = 1; // DIPSW
    //TRISBbits.TRISB7 = 1; // DIPSW

    ANSELB = 0;             // PORTB Digital

    // Setup A/D
    TRISAbits.TRISA0 = 1;   // AN0/RA0 input
    ANSELAbits.ANSA0 = 1;   // Select analog
    ADCON1 = 0b00110000;    // Left justified, RC, VSS-Vref-, VCC-Vref+
    ADCON0bits.CHS = 0;     // Select AN0

    // Pull-up PORTB0
    OPTION_REGbits.nWPUEN = 0;  // enable master pull-up

    // INT0 falling edge sens
    OPTION_REGbits.INTEDG = 0;

    // Setup Timer1
    T1CONbits.TMR1CS = 0;   // Use 1/4 osc
    T1CONbits.T1CKPS = 0;   // Use 1/1 prescaler
    T1CONbits.nT1SYNC = 1;  // No sync
    T1GCONbits.TMR1GE = 0;

    // Enable INT
    INTCONbits.GIE = 1;
    INTCONbits.PEIE = 1;

    // Enable interrupt on change bit for PushSw.
    IOCBNbits.IOCBN3 = 1;

    while(1){

        delay2 = (long)rand() * TXDELAY_MAX / 32768;
        delay1 = SAMPLE_INTERVAL - delay2;

        // Read sensor
        sdata.uldata = read_sensor();
        if(sdata.uldata == SDATA_INVALID){
            // Read sensor failed.
            low_power_sleep(delay1 + delay2, INTERRUPTIBLE);
            continue;
        }
        // Read sensor ok

        // Read id from DIPSW
        xdata.id  = 0;
        xdata.id += (~PORTB & 0x10) ? 8 : 0;
        xdata.id += (~PORTB & 0x20) ? 4 : 0;
        xdata.id += (~PORTB & 0x40) ? 2 : 0;
        xdata.id += (~PORTB & 0x80) ? 1 : 0;

        // If id changed, use different random seed.
        if(xdata.id != id){
            id = xdata.id;
            srand(id);
        }

        // Read battery voltage
        xdata.vbatt = read_vbatt();

        // Fill data
        xdata.data[0]  = (unsigned char) ((sdata.temp >> 4) & 0xff);
        xdata.data[1]  = (unsigned char) ((sdata.temp << 4) & 0xf0);
        xdata.data[1] += (unsigned char) ((sdata.humid >> 8) & 0x0f);
        xdata.data[2]  = (unsigned char)  (sdata.humid & 0xff);

        if(nforce > 0){
            nforce --;
        }

        // Delay 1
        if(nforce == 0){
            if(low_power_sleep(delay1, INTERRUPTIBLE)){
                // Force transmit if PushSw pressed during sleep.
                nforce = 10;
            }
        }

        // Transmit data
        send_data(&xdata);

        // Delay 2
        if(nforce == 0){
            if(low_power_sleep(delay2, INTERRUPTIBLE)){
                // Force transmit if PushSw pressed during sleep.
                nforce = 10;
            }
        }
    }
}
// EOF
